import schedule as baseSchedule
import datetime
import random
from chatqhelper import debug


logger = debug.logger(__name__)


class UTCScheduler(baseSchedule.Scheduler):
    @property
    def idle_seconds(self):
        return (self.next_run - datetime.datetime.utcnow()).total_seconds()

    def every(self, interval=1):
        job = UTCJob(interval, self)
        return job

    def run_pending(self):
        runnable_jobs = (job for job in self.jobs if job.should_run)
        for job in sorted(runnable_jobs):
            try:
                self._run_job(job)
            except Exception as e:
                logger.error('exception while running scheduled jobs -- ' + str(e))
                debug.log_traceback(logger)


class UTCJob(baseSchedule.Job):
    @property
    def should_run(self):
        return datetime.datetime.utcnow() >= self.next_run

    def _schedule_next_run(self):
        assert self.unit in ('seconds', 'minutes', 'hours', 'days', 'weeks')
        if self.latest is not None:
            assert self.latest >= self.interval
            interval = random.randint(self.interval, self.latest)
        else:
            interval = self.interval

        self.period = datetime.timedelta(**{self.unit: interval})
        self.next_run = datetime.datetime.utcnow() + self.period
        if self.start_day is not None:
            assert self.unit == 'weeks'
            weekdays = (
                'monday',
                'tuesday',
                'wednesday',
                'thursday',
                'friday',
                'saturday',
                'sunday'
            )

            assert self.start_day in weekdays
            weekday = weekdays.index(self.start_day)
            days_ahead = weekday - self.next_run.weekday()
            if days_ahead <= 0:
                days_ahead += 7

            self.next_run += datetime.timedelta(days_ahead) - self.period

        if self.at_time is not None:
            assert self.unit in ('days', 'hours') or self.start_day is not None
            kwargs = {
                'minute': self.at_time.minute,
                'second': self.at_time.second,
                'microsecond': 0
            }

            if self.unit == 'days' or self.start_day is not None:
                kwargs['hour'] = self.at_time.hour

            self.next_run = self.next_run.replace(**kwargs)
            if not self.last_run:
                now = datetime.datetime.utcnow()
                if (self.unit == 'days' and self.at_time > now.time() and self.interval == 1):
                    self.next_run = self.next_run - datetime.timedelta(days=1)
                elif self.unit == 'hours' and self.at_time.minute > now.minute:
                    self.next_run = self.next_run - datetime.timedelta(hours=1)

        if self.start_day is not None and self.at_time is not None:
            if (self.next_run - datetime.datetime.utcnow()).days >= 7:
                self.next_run -= self.period

    def cancel(self):
        self.scheduler.cancel_job(self)

    def run(self):
        ret = None
        try:
            ret = self.job_func()
        except Exception as e:
            logger.error('exception while execute job function -- ' + str(e))
            debug.log_traceback(logger)

        self.last_run = datetime.datetime.utcnow()
        self._schedule_next_run()
        return ret


default_scheduler = UTCScheduler()


def every(interval=1):
    return default_scheduler.every(interval)


def tick():
    default_scheduler.run_pending()


def clear(tag=None):
    default_scheduler.clear(tag)


def next_run():
    return default_scheduler.next_run


def idle_seconds():
    return default_scheduler.idle_seconds


def schedule(setting, task):
    every = setting['every']
    if isinstance(every, str):
        job = getattr(default_scheduler.every(), every)
    elif len(every) == 2:
        interval, time_unit = every
        job = getattr(default_scheduler.every(interval), time_unit)
    else:
        from_interval, to_interval, time_unit = every
        job = getattr(default_scheduler.every(from_interval).to(to_interval), time_unit)

    if 'at' in setting:
        job.at(setting['at'])

    job.do(task)
    return job



from chatqhelper import debug, utils
from chatqhelper.mqtt import MqttErrorPublisher
from chatqhelper.slack import SlackErrorNotifier
from chatqhelper.common import constants
import traceback
from functools import wraps


def notify_on_err(stream=None):
    """
    stream: handler object
    """
    def decorator(function):
        @wraps(function)
        def try_cach(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except Exception as e:
                if issubclass(stream, ErrorStreamHandler):
                    stream.handle(e)
                raise e
        return try_cach
    return decorator


class ErrorStreamHandler():
    @classmethod
    def handle(cls, error, client=None):
        pass


class MqttErrorStreamHandler(ErrorStreamHandler):
    # brand new class publisher with new client
    _mqtt_publisher = MqttErrorPublisher()
    @classmethod
    def handle(cls, error):
        error = utils.err_to_dict(error)
        cls._mqtt_publisher.publish(error)


class SlackErrorStreamHandler(ErrorStreamHandler):
    @classmethod
    def handle(cls, error):
        error_msg = utils.err_to_dict(error)
        SlackErrorNotifier.notify(error_msg)



from slackclient import SlackClient
from  chatqhelper import debug
import os, json, time
from functools import wraps
from chatqhelper.mqtt import MqttClient
from chatqhelper.common import constants
from queue import Queue
import datetime

logger = debug.logger("chatqhelper.notifier")

class SlackNotifier():
    """
    requires environment variable SLACK_NOTIFIER_TOKEN
    provides functions:
        - slack_message
    """
    _client = SlackClient(os.environ.get('SLACK_NOTIFIER_TOKEN')) if os.environ.get('SLACK_NOTIFIER_TOKEN') else None
    def __init__(self, token=None):
        # partial singleton, only create a new client if the token is given.
        self._client = SlackClient(token) if token else self.__class__._client
        self._message_queue = []
    
    def notify(self, message, channel, username=''):
        """
        post a message to slack channel
        args:
            - message: string, the text to post
            - channel: string, channel to post the message to
            - username: notifier name, default to service name (hostname of the container)
        """
        sc = self._client
        if not sc:
            logger.warning('invalid slack client or token. message not set')
            return

        res = sc.api_call(
            'chat.postMessage',
            channel=channel, text=message,
            username=username or constants.SERVICE_NAME)
        if not res.get('ok'):
            logger.error(res)

    def attach(self, attachments, channel, username=''):
        """
        post a message to slack channel
        args:
            - message: string, the text to post
            - channel: string, channel to post the message to
            - username: notifier name, default to service name (hostname of the container)
        """
        sc = self._client
        if not sc:
            logger.warning('invalid slack client or token. message not set')
            return

        res = sc.api_call(
            'chat.postMessage',
            channel=channel, attachments=attachments,
            username=username or constants.SERVICE_NAME)
        if not res.get('ok'):
            logger.error(res)


class SlackErrorNotifier():
    INTERNAL_ERR = 'Internal Error'
    CLIENT_ERR = 'Client Error'

    _notifier = SlackNotifier()
    _error_stack = []

    @classmethod
    def notify(cls, error_msg, defer=False):
        if defer:
            cls._error_stack.append(error_msg)
        else:
            cls._notify([error_msg])

    @classmethod
    def defer_notify(cls, error_msg):
        cls.notify(error_msg, defer=True)

    @classmethod
    def _get_error_stack(cls, size=20):
        errors = cls._error_stack[:20]
        cls._error_stack = cls._error_stack[20:]
        return errors

    @staticmethod
    def _error_msg_to_attachment(error_msg):
        code = error_msg.get('code', 500)
        color = 'warning' if 400 <= code < 500 else 'danger'
        source_topic = error_msg.get('source_topic')
        exception = error_msg.get('exception')
        message = error_msg.get('message')
        source_payload = error_msg.get('source_payload')
        tb = error_msg.get('traceback')
        service = error_msg.get('service', '')
        container = error_msg.get('hostname', '')

        attachment = {
            'author_name': service,
            'pretext': exception + ': ' + message,
            'color': color,
            'text': '*payload*\n```{}```\n*traceback*\n```{}```'.format(json.dumps(source_payload, indent=4), tb),
            'footer': container,
            'ts': datetime.datetime.utcnow().timestamp()
        }

        if source_topic:
            attachment.update({
                'title': source_topic,
                'title_link': 'https://chatq.gitbook.io/chatq-development-documentation/chatq/message-bus-topic-paths/chat-project#'+source_topic.replace('/', '-'),
            })

        return attachment

    @classmethod
    def _notify(cls, errors=[]):
        if not errors:
            errors = cls._get_error_stack()

        if not isinstance(errors, list):
            errors = [errors]

        attachments = [cls._error_msg_to_attachment(err) for err in errors]
        if attachments:
            cls._notifier.attach(
                attachments=attachments,
                channel=constants.SLACK_ERR_CHANNEL)



import os
import json
from  chatqhelper import debug
import paho.mqtt.client as mqtt
import uuid
from chatqhelper import exceptions, utils
from chatqhelper.common import constants


logger = debug.logger("chatqhelper.mqtt")


class MqttClient(mqtt.Client):
    def __init__(self,**kwargs):
        self._publish_exception = kwargs.pop('publish_exception', True)
        super(MqttClient, self).__init__(**kwargs)
        self.ignore_topics = set()
        self.is_log_request = False
        if self._publish_exception:
            self._error_publisher = MqttErrorPublisher(mqtt_client=self)

    def ignore(self, topic):
        self.ignore_topics.add(topic)

    def log_request(self, is_log=True):
        self.is_log_request = is_log

    def _handle_on_message(self, message):
        try:
            if message.topic.strip() in self.ignore_topics:
                return
            payload = message.payload
            payload = payload.decode('utf-8') if isinstance(payload, bytes) else payload
            message.payload = json.loads(payload)
            if self.is_log_request:
                logger.info("request: " + message.topic.strip() + " -- " + str(message.payload))

            self._custom_call_callback_list(message)
        except Exception as e:
            self.publish_exception(e, message)

    def _custom_call_callback_list(self, message):
        """
            This method is use to replace the default behavior of _handle_on_message
            We need this class to call not one but a list of callback.
            We also do some custom variable as well. Like split topic out of message...
        """
        with self._callback_mutex:
            try:
                iterator = self._on_message_filtered.iter_match(message.topic)
            except Exception:
                logger.info("TOPIC NOT FOUND")
            else:
                for callback_dict in iterator:
                    for callback in list(callback_dict.values()):
                        with self._in_callback:
                            callback(self, message)

            if self.on_message:
                with self._in_callback:
                    self.on_message(self, message)

    def message_callback_add(self, sub, callback):
        """
            To allow multiple callback on a single topic, we need to do a few custom.
            This method will replace _on_message_filtered map from single callback map
            to a list of callback
        """
        if callback is None or sub is None:
            raise ValueError("sub and callback must both be defined.")

        callback_id = str(uuid.uuid4())
        with self._callback_mutex:
            try:
                self._on_message_filtered[sub][callback_id] = callback
            except Exception:
                self._on_message_filtered[sub] = {
                    callback_id: callback
                }

        return callback_id

    def message_callback_remove(self, callback_id_list):
        """
            Because we now allow multiple call back on a single topic. We need a way to remove it
            using a unique id that is generated during callback registration
        """
        for sub, callback_id in callback_id_list:
            with self._callback_mutex:
                try:
                    del self._on_message_filtered[sub][callback_id]
                except KeyError:  # no such subscription
                    pass

    def publish_exception(self, exception, message):
        logger.error(
            'exception from message: {0} {1}'.format(
                message.topic,
                str(message.payload)
            )
        )

        error = utils.err_to_dict(exception)
        #remove traceback from reply to frontend
        tb = error.pop('traceback', None)

        logger.error(error)
        debug.log_traceback(logger)
        if not self._publish_exception:
            return
        try:
            correlation_id = message.payload.get('correlation_id', None)
            # reply
            if correlation_id is not None:
                self.reply(message, error)

            # publish to exceptions/ topics
            if self._error_publisher:
                source_payload = message.payload
                if 'auth_key' in source_payload:
                    source_payload['auth_key'] = '---'
                error.update({
                    'source_topic': message.topic,
                    'source_payload': message.payload,
                    # add back traceback
                    'traceback': tb
                })
                self._error_publisher.publish(error)
        except Exception as e:
            # nothing we can do here except for logging
            logger.error(e)
            debug.log_traceback(logger)

    def reply(self, msg, data, log_response=False):
        topic = msg.topic + "/reply-to/" + str(msg.payload.get('correlation_id'))
        payload = json.dumps(data)
        if log_response:
            logger.info("reply: " + topic + " -- " + payload)

        self.publish(topic, payload)

    def setup_message_callbacks(self, *args):
        for topic, callback in args:
            self.message_callback_add(topic, callback)

    @classmethod
    def create(cls, on_connect, is_log_request=False, **kwargs):
        if not 'post_fix' in kwargs:
            kwargs['post_fix'] = 'main'
        post_fix = kwargs.pop('post_fix')
        post_fix = '-' + post_fix if not post_fix[0] == '-' else post_fix
        kwargs['post_fix'] = post_fix

        if 'client_id' not in kwargs:
            kwargs['client_id'] = constants.HOSTNAME + kwargs.pop('post_fix','')+ '-pid{}'.format(str(os.getpid()))
            logger.info('Chatq MqttClient initalized with default client id {}'.format(kwargs['client_id']))

        client = cls(**kwargs)
        client.log_request(is_log_request)
        client.on_connect = on_connect
        client.username_pw_set(
            username=os.environ.get('SOL_USERNAME', ''),
            password=os.environ.get('SOL_PASSWORD', '')
        )

        client.connect_async(
            os.environ.get('SOL_URI', ''),
            int(os.environ.get('SOL_MQTT_PORT', '0')),
            60
        )

        return client


class MqttErrorPublisher():
    """ error publisher will publish error to exceptions/backend
    """
    def __init__(self, mqtt_client=None):
        self._mqtt_client = mqtt_client
        if not self._mqtt_client:
            mqtt_client = MqttClient.create(self._on_connect, post_fix='err-publisher')
            mqtt_client.loop_start()

    def _on_connect(self, client, userdata, flags, rc):
        print("connect with code " + str(rc))
        self._mqtt_client = client

    def publish(self, error_msg):
        """
        publish error with added info
         - trace_back
         - service
         - hostname
        """
        client = self._mqtt_client
        if client:
            client.publish(
                'exceptions/backend',
                json.dumps(error_msg)
            )
        else:
            logger.warning('invalid mqtt client, not publishing error')



"""
The module suppporting getting records from google sheet and perform filter and projection.
To use google sheet api 2 main steps must be done:
    - enable google sheet api and get the service account credentials
    - share the sheet with the associated email in service account credentials
"""
import json
from oauth2client.service_account import ServiceAccountCredentials
import gspread
import os, copy

SCOPES = 'https://www.googleapis.com/auth/spreadsheets'
SERVICE_SECRET_FILE =  os.environ['GOOGLE_APPLICATION_CREDENTIALS']


BOOL_TRUE = 'TRUE'
BOOL_FALSE = 'FALSE'


def get_client():
    """
    get a new google sheet client and login
    require service file containg google service credentials
    """
    credentials = ServiceAccountCredentials.from_json_keyfile_name(SERVICE_SECRET_FILE, SCOPES)
    client = gspread.Client(credentials)
    client.login()
    return client

class Record():
    def __init__(self, records=[], spreadsheet=None, worksheet=None, client=None):
        self._records = records
        self._record_keys = records[0].keys() if records else []
        self._spreadsheet = spreadsheet
        self._worksheet = worksheet
        self._client = client
    
    def filter(self, filt={}):
        """
        filter should be 1 level dictionary.
        to do nested filter operation, perform 2 separated filter, one after another
        sample filter:
        filter = {
            "Index": {
                LTE: 50
            }
        }
        """
        if not self._records or not filt:
            return self._records

        rec_keys = self._record_keys
        filt_keys = filt.keys()

        # validate filter keys. warn if filter keys not in record keys
        missing_keys = list(set(filt_keys) - set(rec_keys))
        if missing_keys:
            raise Exception('could not filter. Missing key in record {}'.format(missing_keys))

        results = []
        for record in self._records:
            results.append(record)
            for key in filt_keys:
                opt_code, param = get_operator_param(filt[key])
                if not operator_filter[opt_code](record[key], param):
                    results.pop()
                    break
        
        self._records = results
        return self

    def project(self, proj):
        """
        The main purpose of projection is to reduce memory foodprint when it deletes fields/keys not needed by applications
        Projection should be a dictionary with keys as the fields to be projected and values is true/false or 0/1
        Projection can be a list of desired fields
        """
        if not self._records or not proj:
            return self._records
        
        # if it's a list, then map to key:True
        if isinstance(proj, list):
            proj = {k:True for k in proj}

        rec_keys = self._record_keys
        yes_keys = [k for k in proj.keys() if proj[k]]
        yes_keys = yes_keys if yes_keys else copy.deepcopy(list(rec_keys))
        no_keys = [k for k in proj.keys() if not proj[k]]
        yes_keys = list(set(yes_keys) - set(no_keys))

        if not yes_keys:
            raise Exception('invalid projection, nothing left to show')
        
        results = [
            {
                k:record[k] for k in yes_keys
            } for record in self._records
        ]

        self._records = results
        self._record_keys = results[0].keys()
    
    def to_json(self):
        return copy.deepcopy(self._records)

    def reload(self):
        records = self._worksheet.get_all_records()
        self._records = records

    @classmethod
    def get_all_records_from_sheet(cls, spreadsheet_id, worksheet_name, client=None):
        if not client:
            client = get_client()
        spreadsheet = client.open_by_key(spreadsheet_id)
        worksheet = spreadsheet.worksheet(worksheet_name)
        records = worksheet.get_all_records()
        records = Record(records, spreadsheet, worksheet, client)
        return records

# ------------------------------------------------------

# operators
EQ = '$EQ'
NE = '$NE'
NOT = '$NOT'
IN = '$IN'
NOTIN = '$NOTIN'
HAS = '$HAS'
NOTHAS = '$NOTHAS'
GT = '$GT'
GTE = '$GTE'
LT = '$LT'
LTE = '$LTE'

# operator functions
def eq_filt(value, param):
    return value == param

def ne_filt(value, param):
    return value != param

def in_filt(value, param):
    return value in param

def notin_filt(value, param):
    return value not in param

def has_filt(value, param):
    return param in value

def nothas_filt(value, param):
    return param not in value

def gt_filt(value, param):
    return type(param)(value) > param

def gte_filt(value, param):
    return type(param)(value) >= param

def lt_filt(value, param):
    return type(param)(value) < param

def lte_filt(value, param):
    return type(param)(value) <= param

def get_operator_param(filt):
    operator, param = EQ, filt
    if isinstance(filt, dict):
        operator = list(filt)[0]
        param = filt[operator]
    return operator, param

# a map of operator name to operator function
operator_filter = {
    EQ: eq_filt,
    NE: ne_filt,
    IN: in_filt,
    NOTIN: notin_filt,
    HAS: has_filt,
    NOTHAS: nothas_filt,
    GT: gt_filt,
    GTE: gte_filt,
    LT: lt_filt,
    LTE: lte_filt,
}

#----------------------------------------------------------

if __name__ == "__main__":
    pass
#     # print(opt_parm({NOT:{ISIN:['a','b','c']}}))
#     print(get_records({USER_FN:{NOT:{ISIN:["Khoa", "Ilya"]}}},{USER_FN:1, USER_ID:0}))


class BaseException(Exception):
    def __init__(self, message, code):
        self.code = code
        super(BaseException, self).__init__(message)


class ClientError(BaseException):
    def __init__(self, message="", ext=0):
        code = 400 + ext
        super(ClientError, self).__init__(message, code)


class ServerError(BaseException):
    def __init__(self, message="", ext=0):
        code = 500 + ext
        super(ServerError, self).__init__(message, code)



from setuptools import setup, find_packages
import os

setup(
      name='chatqhelper',
      version=os.environ['RELEASE_VERSION'],
      description='chatqhelper',
      url='https://github.com/ChatQSG/chatq-helping-hand',
      author='ChatQ',
      author_email='',
      packages=find_packages(),
      install_requires=[
            'paho-mqtt==1.3.1',
            'schedule==0.5.0',
            'gspread==3.0.1',
            'google-api-python-client==1.7.4',
            'oauth2client==4.1.3',
            'slackclient==1.3.0'
      ],
      zip_safe=False
)


import dateutil.parser
import datetime
import traceback
from chatqhelper.common import constants


def chunks(items, size):
    for i in range(0, len(items), size):
        yield items[i:i + size]


def str_to_ts(timestr):
    return dateutil.parser.parse(timestr).replace(tzinfo=datetime.timezone.utc).timestamp()

def err_to_dict(err):
    return {
        'exception': str(err.__class__.__name__),
        'message': str(err),
        'code': getattr(err, 'code', 500),
        'traceback': traceback.format_exc(),
        'service': constants.SERVICE_NAME,
        'hostname': constants.HOSTNAME
        }
